#ifndef bqt_dyn_h
#define bqt_dyn_h

#include "autoptr"
#include <cmath>
#include <iostream>
#include <list>

#define DYNDEBUG 0

/* Expression identifiers. Used for optimizing. */
typedef enum dyntype
{
    DYNTYPE_VALUE,
    DYNTYPE_PLUSMINUS,
    DYNTYPE_TIMESDIV,
    DYNTYPE_MOD,
    DYNTYPE_FUNC
} dyntype;

/* Throwable type: error code. */
class dynerror
{
public:
    enum type { getunsetval=1, getunsetexp, virtualcall };
private:
    type typi;
public:
    explicit dynerror(type t) : typi(t) { }
    operator const char * () const
    {
        return typi==getunsetval ? "Tried to get a value of UNSET variable"
             : typi==getunsetexp ? "Tried to get expression of UNSET variable"
             : typi==virtualcall ? "Pure virtual function called"
             : "Failing dyn error";
    }
};

std::ostream &operator<< (std::ostream &dest, dyntype t)
{
    switch(t)
    {
        case DYNTYPE_VALUE: dest << "VALUE"; break;
        case DYNTYPE_PLUSMINUS: dest << "PLUSMINUS"; break;
        case DYNTYPE_TIMESDIV: dest << "TIMESDIV"; break;
        case DYNTYPE_MOD: dest << "MOD"; break;
        case DYNTYPE_FUNC: dest << "FUNC"; break;
        default: dest << '?' << (int)t << '?';
    }
    return dest;
}

/* This is the base type for stored expressions.
 *
 */
template<typename T>
class dyni : public ptrable
{
public:
    virtual ~dyni()
    {
#if DYNDEBUG > 0
        std::cout << "dyni dying: \"" << std::flush;
        debug(); std::cout << '"' << std::endl;
#endif
    }
    virtual const T getvalue() const = 0;
    virtual void debug() const = 0; //{ throw dynerror(dynerror::virtualcall); }
    virtual const dyntype gettype() const = 0;
};

/* This is the dynamic expression class.
 * It contains a pointer to the actual expression, which
 * may be shared by multiple dyn-instances. The expression
 * contains a reference counter, and is automatically
 * destructed when the last dyn pointing to it dies.
 */
template<typename T>
class dyn : public ptrable
{
protected:
    autoptr<dyni<T> > v;
public:
    explicit dyn(dyni<T> *p = NULL);
    dyn(const dyn<T> &);
    dyn(const T &);
    virtual ~dyn()
    {
#if DYNDEBUG > 0
        std::cout << "dyn dying: \"" << std::flush;
        debug(); std::cout << '"' << std::endl;
#endif
    }
    
    const T getvalue() const
    {
        if(!v) throw dynerror(dynerror::getunsetval);
        return v->getvalue();
    }
    void debug() const
    {
        if(!v) throw dynerror(dynerror::getunsetval);
        v->debug();
    }
    dyn &operator= (const dyn<T> &p);
    dyn &operator= (dyni<T> *p);
    
    dyn &operator= (const T &value); // constant
    
    /* typecast operator */
    //operator T () const { return getvalue(); }
    
    const dyni<T> *const getexpr() const
    {
        if(!v) throw dynerror(dynerror::getunsetexp);
        return v;
    }
};

/* Expression type: Constant */
template<typename T>
class dynvalue : public dyni<T>
{
    T v;
public:
    explicit dynvalue(const T &c) : v(c) { }
    virtual const T getvalue() const
    {
        return v;
    }
    virtual void debug() const { std::cout << v << std::flush; }
    virtual const dyntype gettype() const { return DYNTYPE_VALUE; }
};

template<typename T>
dyn<T>::dyn(dyni<T> *p = NULL) : v(p)
{
#if DYNDEBUG > 0
    if(p){std::cout << "dyn explicitly created: \"" << std::flush;
    debug(); std::cout << "\" (" << v->gettype() << ')' << std::endl;}
#endif
}
template<typename T> dyn<T>::dyn(const dyn<T> &c) :ptrable(), v(c.v) { }
template<typename T> dyn<T>::dyn(const T &c) : v(new dynvalue<T>(c)) { }
template<typename T> dyn<T> &dyn<T>::operator= (dyni<T> *p) { v = p; return *this; }
template<typename T> dyn<T> &dyn<T>::operator= (const dyn<T> &p) { v = p.v; return *this; }
template<typename T> dyn<T> &dyn<T>::operator= (const T &value) { return *this = new dynvalue<T> (value); }

template<typename T>
class dyn2pair : public ptrable
{
private:
    dyn2pair<T> &operator= (const dyn2pair<T> &) { }
#if DYNDEBUG > 0
    void debuggi(const char *s)
    {
        std::cout << s << "ed elem: " << (nega?"-/ " : "+* ");
        expr.debug();
        std::cout << std::endl;
    }
#endif
public:
    bool nega;
    dyn<T> &expr;
public:
    dyn2pair(bool neg, dyn<T> &exp) : nega(neg), expr(exp)
    {
#if DYNDEBUG > 0
    	debuggi("Creat");
#endif
    }
    virtual ~dyn2pair()
    {
#if DYNDEBUG > 0
    	debuggi("Kill");
#endif
    }
};

template<typename T>
class dyntmpkeep
{
private:
    typedef autoptr<dyn<T> > tmptype;
    std::list<tmptype> tmplist;
public:
    void addtmp(dyn<T> *p)
    {
#if DYNDEBUG > 0
    	std::cout << "Added temporary: ";
    	p->debug();
    	std::cout << std::endl;
#endif
        /* Add p to the automatic temporary variable deletion list 
         * (They will be deleted when the destructor is called, due to autoptr.
         */
        tmplist.push_back(p);
    }
    dyntmpkeep()
    {
#if DYNDEBUG > 0
    	std::cout << "Created empty temporary store\n";
#endif
    }
    virtual ~dyntmpkeep()
    {
#if DYNDEBUG > 0
    	if(!tmplist.size())
	    	std::cout << "Deleting empty temporary store\n";
	    else
	    {
	    	std::cout << "Deleting temporaries\n";
	    	std::list<tmptype>::const_iterator i;
	    	for(i=tmplist.begin(); i!=tmplist.end(); ++i)
	    	{
	    		std::cout << ": ";
	    		(*i)->debug();
	    		std::cout << std::endl;
	    	}
	    	std::cout << "[end of list]\n";
	    }
#endif
    }
};

template<typename T>
class dyn2coll : public dyni<T>, public dyntmpkeep<T>
{
private:
    char sgneg, sgpos;
    typedef autoptr<dyn2pair<T> > elemtype;
    typedef std::list<elemtype> elemlist;
protected:
    elemlist elems;
public:
    dyn2coll(char n, char p) : sgneg(n), sgpos(p)
    {
#if DYNDEBUG > 1
        std::cout << "Created dyn2coll('" << n << "', '" << p << "')\n";
#endif
    }
    virtual ~dyn2coll()
    {
#if DYNDEBUG > 1
        std::cout << "Killing dyn2coll \"" << std::flush;
        debug();
        std::cout << '"' << std::endl;
#endif
    }
    virtual void debug() const;
    void add(dyn<T> &elem, bool nega);
    void add(dyn<T> *elem, bool nega);
};

template<typename T>
void dyn2coll<T>::debug() const
{
    if(!elems.size()) { std::cout << "[EMPTY]"; return; }
    if(elems.size() > 1)std::cout << '(';
    elemlist::const_iterator i = elems.begin();
    (*i)->expr.debug();
    while(++i != elems.end())
    {
        std::cout << (char)((*i)->nega ? sgneg : sgpos);
        (*i)->expr.debug();
    }
    if(elems.size() > 1)std::cout << ')';
}

template<typename T>
void dyn2coll<T>::add(dyn<T> *elem, bool nega)
{
	addtmp(elem);
	add(*elem, nega);
}

template<typename T>
void dyn2coll<T>::add(dyn<T> &elem, bool nega)
{
    #if 1
    const dyni<T> *expr = elem.getexpr();
#if DYNDEBUG > 2
    std::cout << "ADD " << gettype() << ": "; debug(); std::cout << " <- " << expr->gettype() << ": "; expr->debug(); std::cout << std::endl;
#endif
    if(expr->gettype() == gettype())
    { 
        /* For unknown reason, dynamic_cast causes segmentation fault.
         * static_cast does not, and works as supposed.
         */
        const dyn2coll<T> *expc = static_cast<const dyn2coll<T> *> (expr);
        const elemlist e = expc->elems;
        elemlist::const_iterator i;
        for(i=e.begin(); i!=e.end(); ++i)
            elems.push_back(new dyn2pair<T> (nega ^ (*i)->nega, (*i)->expr));
    }
    else
    #endif
        elems.push_back(new dyn2pair<T> (nega, elem));
    //std::cout << "DONE" << std::endl;
}


template<typename T>
class dynplusminus : public dyn2coll<T>
{
public:
    dynplusminus() : dyn2coll<T> ('-', '+') { }
    virtual const T getvalue() const;
    virtual const dyntype gettype() const { return DYNTYPE_PLUSMINUS; }
};

template<typename T>
class dyntimesdiv : public dyn2coll<T>
{
public:
    dyntimesdiv() : dyn2coll<T> ('/', '*') { }
    virtual const T getvalue() const;
    virtual const dyntype gettype() const { return DYNTYPE_TIMESDIV; }
};

template<typename T>
const T dynplusminus<T>::getvalue() const
{
    T result = 0;
    for(elemlist::const_iterator i = elems.begin(); i != elems.end(); ++i)
        if((*i)->nega) result -= (*i)->expr.getvalue();
        else result += (*i)->expr.getvalue();
    return result;
}

template<typename T>
const T dyntimesdiv<T>::getvalue() const
{
    T result = 1;
    for(elemlist::const_iterator i = elems.begin(); i != elems.end(); ++i)
        if((*i)->nega) result /= (*i)->expr.getvalue();
        else result *= (*i)->expr.getvalue();
    return result;
}

#define dyn2opdef(op, nega, obj) \
  template<typename T> \
  const dyn<T> operator op (const dyn<T> &l, const dyn<T> &r) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(new dyn<T> (l), false); \
    p->add(new dyn<T> (r), nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator op (const dyn<T> &l, dyn<T> &r) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(new dyn<T> (l), false); \
    p->add(r, nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator op (dyn<T> &l, const dyn<T> &r) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(l, false); \
    p->add(new dyn<T> (r), nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator op (dyn<T> &l, dyn<T> &r) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(l, false); \
    p->add(r, nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator op (const dyn<T> &l, const T &value) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(new dyn<T> (l), false); \
    p->add(new dyn<T> (value), nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator op (dyn<T> &l, const T &value) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(l, false); \
    p->add(new dyn<T> (value), nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator op (const T &value, const dyn<T> &r) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(new dyn<T> (value), false); \
    p->add(new dyn<T> (r), nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator op (const T &value, dyn<T> &r) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(new dyn<T> (value), false); \
    p->add(r, nega); \
    return dyn<T> (p); \
  }

dyn2opdef(+, false, dynplusminus)
dyn2opdef(-, true,  dynplusminus)
dyn2opdef(*, false, dyntimesdiv)
dyn2opdef(/, true,  dyntimesdiv)

#undef dyn2opdef

#define dyn2opdef(opa, nega, obj) \
  template<typename T> \
  const dyn<T> &operator opa (dyn<T> &dest, const dyn<T> &r) \
  { \
    dyn<T> *tmp = new dyn<T> (dest); \
    obj<T> *p = new obj<T>; \
    p->add(tmp, false); \
    p->add(new dyn<T> (r), nega); \
    return (dest = p); \
  } \
  template<typename T> \
  const dyn<T> &operator opa (dyn<T> &dest, dyn<T> &r) \
  { \
    dyn<T> *tmp = new dyn<T> (dest); \
    obj<T> *p = new obj<T>; \
    p->add(tmp, false); \
    p->add(r, nega); \
    return (dest = p); \
  } \
  template<typename T> \
  const dyn<T> &operator opa (dyn<T> &dest, const T &value) \
  { \
    dyn<T> *tmp = new dyn<T> (dest); \
    obj<T> *p = new obj<T>; \
    p->add(tmp, false); \
    p->add(new dyn<T> (value), nega); \
    return (dest = p); \
  }

dyn2opdef(+=, false, dynplusminus)
dyn2opdef(-=, true,  dynplusminus)
dyn2opdef(*=, false, dyntimesdiv)
dyn2opdef(/=, true,  dyntimesdiv)

#undef dyn2opdef

#if 1 /* funmod */

/* Expression base type: Modulo */
template<typename T>
class dynmod : public dyni<T>, public dyntmpkeep<T>
{
private:
    dyn<T> &L, &R;
public:
    dynmod(const dyn<T> &l, const dyn<T> &r) : L(*new dyn<T> (l)), R(*new dyn<T> (r)) { addtmp(&L); addtmp(&R); }
    dynmod(const dyn<T> &l, dyn<T> &r) : L(*new dyn<T> (l)), R(r) { addtmp(&L); }
    dynmod(dyn<T> &l, const dyn<T> &r) : L(l), R(*new dyn<T> (r)) { addtmp(&R); }
    dynmod(dyn<T> &l, dyn<T> &r) : L(l), R(r) {}
    virtual const T getvalue() const { return L.getvalue() % R.getvalue(); }
    virtual void debug() const { std::cout << '{'; L.debug();
                                 std::cout << '%'; R.debug();
                                 std::cout << '}'; }
    virtual const dyntype gettype() const { return DYNTYPE_MOD; }
};
template<typename T, typename T2, typename T3>
const dyn<T> operator % (T2 l, T3 r)
{
    return dyn<T> (new dynmod<T> (l, r));
}
template<typename T, typename T2>
const dyn<T> &operator %= (dyn<T> &dest, T2 v)
{
    dyn<T> *tmp = new dyn<T> (dest);
    dynmod<T> *p = new dynmod<T> (*tmp, v);
    p->addtmp(tmp);
    retunr (dest = p);
}

template<typename T>
class dynfun : public dyni<T>, dyntmpkeep<T>
{
private:
    dyn<T> &E;
    T (*const F) (T);
    const char *funame;
public:
    dynfun(const dyn<T> &exp, T (*const fun) (T), const char *f);
    dynfun(dyn<T> &exp, T (*const fun) (T), const char *f);
    virtual ~dynfun();
    virtual const T getvalue() const { return F(E.getvalue()); }
    virtual void debug() const;
    virtual const dyntype gettype() const { return DYNTYPE_FUNC; }
};
template<typename T>
dynfun<T>::dynfun(const dyn<T> &exp, T (*const fun) (T),  const char *f)
             : E(*new dyn<T> (exp)), F(fun), funame(f)
{
	addtmp(&E);
#if DYNDEBUG > 0
	std::cout << "Created dynfun(const "; exp.debug();
	std::cout << ", \"" << f << "\")\n";
#endif
}
template<typename T>
dynfun<T>::dynfun(dyn<T> &exp, T (*const fun) (T), const char *f)
             : E(exp), F(fun), funame(f)
{
#if DYNDEBUG > 0
	std::cout << "Created dynfun("; exp.debug();
	std::cout << ", \"" << f << "\")\n";
#endif
}
template<typename T>
dynfun<T>::~dynfun()
{
#if DYNDEBUG > 0
	std::cout << "Killed dynfun: ";
	E.debug();
	std::cout << std::endl;
#endif
}

template<typename T>
void dynfun<T>::debug() const
{
    std::cout << funame << '('; E.debug();
    std::cout << ')';
}
#define dynfundef(fu) \
template<typename T> \
inline const dyn<T> fu(const dyn<T> &e) \
{ \
    return dyn<T> (new dynfun<T> (e, std::fu, #fu)); \
} \
template<typename T> \
inline const dyn<T> fu(dyn<T> &e) \
{ \
    return dyn<T> (new dynfun<T> (e, std::fu, #fu)); \
}

dynfundef(acos)
dynfundef(asin)
dynfundef(atan)
dynfundef(ceil)
dynfundef(cos)
dynfundef(cosh)
dynfundef(exp)
#if defined __USE_MISC || defined __USE_XOPEN_EXTENDED || defined __USE_ISOC99
dynfundef(expm1)
dynfundef(log1p)
dynfundef(logb)
dynfundef(cbrt)
#endif
#ifdef __USE_ISOC99
dynfundef(exp2)
dynfundef(log2)
dynfundef(round)
#endif
dynfundef(floor)
dynfundef(log)
dynfundef(log10)
dynfundef(sin)
#ifdef __USE_GNU
dynfundef(sincos)
#endif
#ifdef __USE_MISC
dynfundef(significand)
#endif
dynfundef(sinh)
dynfundef(sqrt)
dynfundef(tan)
dynfundef(tanh)

#undef dynfundef

#endif /* funmod */

#endif
